/*
 * _Beep_dsp_handling.h
 *
 *  Created on: 8. jan. 2022
 *      Author: oyvindteig
 */


#ifndef DSP_HANDLING_H_
#define DSP_HANDLING_H_

    typedef struct {
        unsigned     num_samples; // The number of valid samples
        bool         is_full;
        unsigned     num_lost_overflowed;
        bool         was_late;
    } mic_samples_info_t;

    // The buffer used to be part of the above struct, but with the cast to complex array it looks better alone.
    // Both need to be sent (in send_pair) per "batch", but the overhead is about nought on this processor..
    //
    typedef struct {
        int64_t dummy_avoid_memory_error_align_below_at_double_word; // or make the array file global in the file
        union {
            // ================================================
            // Reuse of buffer, avoids ram usage and memcpy
            // with union instead of complex casts to pointers
            // ================================================

            // All these unions must have the same size. There is an
            //     xassert ((sizeof (this struct) == NUM_MIC_SAMPLES * 4));
            // in the code. There also check by the compiler at the same place

            // Fill samples into:
            //
            mic_sample_t mic_samples [NUM_MIC_SAMPLES];

            // For:
            //     dsp_fft_bit_reverse
            //     dsp_fft_forward
            //     dsp_fft_split_spectrum
            // Instead of this (which seemed to work):
            //     dsp_complex_t (* restrict complex_array_ptr) = (dsp_complex_t(*)) buffer.u.mic_samples;
            // This cast is explained in https://www.katjaas.nl/realFFT/realFFT2.html
            // Plus in lib_dsp.pdf
            //
            dsp_complex_t complex_array [NUM_FFT_POINTS_COMPLEX];

            // After dsp_fft_split_spectrum

            // Instead of this:
            //     dsp_complex_t (* restrict w)[2] = (dsp_complex_t (*)[2])pts;
            // spelt out as (which I did have problems with the error size of array too large for the usage in my_dsp_magnitude):
            //     dsp_complex_t  (* restrict freq_spectrum)[NUM_FREQ_SPECTRA] =
            //         (dsp_complex_t (*)[NUM_FREQ_SPECTRA])buffer.u.mic_samples;
            //
            dsp_complex_t complex_half_spectra [NUM_FREQ_SPECTRA][NUM_FREQ_POINTS_REAL];
        } u;
    } mic_samples_buff_t;

    typedef bool get_next_mic_frame_t;

    #define SPECTRUM_Q_FORMAT 24
    typedef q8_24             spectrum_val_t;

    typedef struct {
        uint32_t val_at_ix; // Init is ZERO_VAL
        unsigned ix;
        unsigned val_new_cnt;    // Interesting
        unsigned val_stable_cnt; // Interesting
        unsigned val_aux_cnt;    // Interesting
    } uint_max_value_t; // Max is (2exp32)-1 = UINT_MAX 4294967295

    typedef struct {
        int32_t  val_at_ix; // Init is INT_MIN (limits.h)
        bool     curve_is_all_zeros;
        unsigned ix;
        unsigned val_new_cnt;    // Interesting
        unsigned val_stable_cnt; // Interesting
        unsigned val_aux_cnt;    // Interesting
    } int_max_value_t; // Max is (2exp31)-1 = INT_MAX 2147483647

    typedef struct {
        int32_t  val_at_ix; // Init is INT_MAX (limits.h)
        unsigned ix;
        unsigned val_new_cnt;    // Interesting
        unsigned val_stable_cnt; // Interesting
        unsigned val_aux_cnt;    // Interesting
    } int_min_value_t; // Min is -2exp31 = INT_MIN −2147483648

    typedef struct {
        int_max_value_t freq_max_value;
    } data_ch_results_t;

    typedef enum {
        in_ui_in_ack,            // then in_ui
        out_dsp_result_after_ack // then out_dsp_result
    } knock_reason_e;

    typedef enum {
        in_ui,         // after in_ui_in_ack
        out_dsp_result // after out_dsp_result_after_ack
    } ack_reason_e;

    typedef struct { // MAX_SIZEOF_NONBLOCKING_STREAMING_CHAN
        knock_reason_e knock_reason;
    } data_sch_knock_t;

    typedef struct {
        ack_reason_e ack_reason;
    } data_sch_ack_t;

    // To filter noise from button
    // See "2022 05 18 v0820 PORTS_1D_1N_AT_5 Noise from buttons at 40 dB SDS00083.png"
    // 64 ms before this goes high to 64 ms after it's gone low: discard results
    //
    typedef struct {
        // I tried first with an enum with 4 values, then with two booleans, but the problem was
        // to make sure that the transient state of the buttons and actually using that state was
        // difficult to get right. So I dropped any static state and now do to transient state
        // only: noise from the buttons when they are pressed or released, but not when held depressed
        //
        unsigned cnt; // Transient state, any count between testing, as long as it counts
    } any_button_any_event_t;

    typedef struct {
        data_sch_ack_t         data_sch_ack;
        any_button_any_event_t any_button_any_event;
    } ch_data_ack_ui_t; // "ui" = "user interface"

    typedef struct {
        spectrum_val_t spectrum [NUM_FREQ_POINTS_REAL];
    } data_ch_logger_t;

    typedef struct {
        unsigned        iof_spectrum;
        spectrum_val_t  spectrum [NUM_FREQ_POINTS_REAL];
        int_max_value_t freq_max_value;
    } ch_data_logger_spectrum_t;

    void buffer_task_x (
            chanend                 ch_in,
            chanend                 ch_yx,
            client port_commands_if if_port_commands);

    // [[combinable]] // Not with ordered select
    void dsp_task_y (
            chanend           ch_out_in,
            chanend           ch_logger_in_3out,
            streaming chanend sch_yb);

    #ifndef DEBUG_DSP_ERROR_LOG
        #error include of _globals.h missing
    #endif

#else
    #error NESTED .h FILE
#endif
