/*
 * _Beep_dsp_handling.xc
 *
 *  Created on: 8. jan. 2022
 *      Author: oyvindteig
 */


#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <xscope.h>   // xScope logging
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <i2c.h>
    #include <i2s.h>
    #include <dsp_qformat.h>
    #include <dsp.h>
    #include <dsp_fft.h>
    #include <dsp_math.h>
    #include <math.h> // log10 etc
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "maths.h"
    #include "mic_array.h"
    #include "maths_fix.h" // Needs stdint.h
    #include "mics_in_headset_out.h"
    #include "_Beep_dsp_params.h"
    #include "param.h"
    #include "i2c_client_task.h"
    #include "port_tile0_task.h"
    #include "logger.h"
    #include "_Beep_dsp_handling.h"

    #define DO_DEBUG_PRINT         0 // debug_print 0 has consequences if (USE_XSCOPE == 1) and amy timing at all needs to be met
    #define DO_DEBUG_PRINT_SPECIAL 1 // debug_print_special
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif

void mic_samples_info_init (mic_samples_info_t &mic_samples_info) {

    mic_samples_info.num_samples         = 0;
    mic_samples_info.num_lost_overflowed = 0;
    mic_samples_info.is_full             = false;
    mic_samples_info.was_late            = false;
} // mic_samples_info_init


bool mic_samples_info_fill_buffer (
        const data_ch_phase3_t phase3_ch_mic_samples, // in
        mic_samples_buff_t     &buffer,
        mic_samples_info_t     &mic_samples_info)      // out NUM_MIC_SAMPLES
{
    bool           is_full_pre         = mic_samples_info.is_full;
    const unsigned num_new_samples     = phase3_ch_mic_samples.samples.rec_max_msg.nums.num_samples;
    unsigned       num_new_samples_valid;

    // Add from incoming:
    mic_samples_info.num_lost_overflowed += phase3_ch_mic_samples.samples.rec_max_msg.nums.num_lost_overflowed;

    if (mic_samples_info.is_full) {

        mic_samples_info.num_lost_overflowed += num_new_samples; // Count and count..
        num_new_samples_valid = 0; // Don't use any

    } else if ((mic_samples_info.num_samples + num_new_samples) >= NUM_MIC_SAMPLES) { // v0618 fix (was >)

        mic_samples_info.is_full = true;
        num_new_samples_valid   = NUM_MIC_SAMPLES - mic_samples_info.num_samples; // Not to overflow array

    } else {

        num_new_samples_valid = num_new_samples; // Cannot overflow array
    }

    #define DO_MEMFILL_TYPE MEMFILL_TYPE_FOR_LOOP

    #if (DO_MEMFILL_TYPE == MEMFILL_TYPE_FOR_LOOP)
        for (unsigned ix=0; ix < num_new_samples_valid; ix++) {
            buffer.u.mic_samples[mic_samples_info.num_samples+ix] =
                    phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[ix];
        }
    #else
        #error
    #endif

    mic_samples_info.num_samples += num_new_samples_valid; // Max is then NUM_MIC_SAMPLES

    xassert (mic_samples_info.num_samples <= NUM_MIC_SAMPLES);

    return (mic_samples_info.is_full != is_full_pre);
} // mic_samples_info_fill_buffer


void send_twice_in_x (
        chanend            ch_out,
        mic_samples_info_t mic_samples_info,
        mic_samples_buff_t buffer)
{
    // send_twice_in_x: trio (to be received on ch_yx in receive_twice_in_y)
    ch_out <: mic_samples_info;
    ch_out <: buffer;
}


void receive_twice_in_y (
        chanend            ch_out,
        mic_samples_info_t &mic_samples_info,
        mic_samples_buff_t &buffer)
{
    // receive_twice_in_y (was sent on ch_yx in send_twice_in_x)
    ch_out :> mic_samples_info;
    ch_out :> buffer;
}


// Alternative: movable pointer, see 14 Double buffering example of the
//             XMOS Programming Guide. 2015/9/18 XMOS (XM004440A).

void buffer_task_x (
        chanend                 ch_in,            // data_ch_phase0_t  then data_ch_phase3_t
        chanend                 ch_yx,            // mic_samples_info_t then get_next_mic_frame_t)
        client port_commands_if if_port_commands)
{
    // Double buffering.
    // One buffer used to be filled up by any sample-num size at a time,
    // while the filled up is waiting to get sent away as a complete
    // frame (on get_next_mic_frame)
    // Nothing to do with DO_ONE_FRAME_DELAY
    //
    #define NUM_BUFFERS 2

    debug_print ("%s\n", "buffer_task_x started");

    data_sch_phase1_t   phase1_sch_mic_sample;
    // data_ch_phase2_t phase2_ch_ack_etc; // No PHASE2
    data_ch_phase3_t    phase3_ch_mic_samples;

    mic_samples_buff_t  buffer            [NUM_BUFFERS]; // The full one is a "frame"
    mic_samples_info_t  mic_samples_info  [NUM_BUFFERS];

    unsigned             iof_buffer_to_fill = 0;
    unsigned             iof_buffer_to_send_late = 0;
    get_next_mic_frame_t get_next_mic_frame = false;
    bool                 ready_to_send_late = false;

    unsigned debug_one_sample_cnt  = 0;
    unsigned debug_all_samples_cnt = 0;

    #if (DEBUG_MIC_SAMPLES_INCREASING > 0)
        mic_sample_t debug_sliding_mic_sample = COUNTING_MIC_SAMPLE_INIT;
    #endif

    for (unsigned ix=0; ix<NUM_BUFFERS; ix++) {
        mic_samples_info_init (mic_samples_info[ix]);
    }

    if_port_commands.write_port (PORT_1D_1N, port_low);

    while (1) {
        [[ordered]]
        select {
            // In buffer_task_x
            case ch_yx :> get_next_mic_frame : { // value true. In buffer_task_x, was sent on ch_yx from dsp_task_y
                // One query, one answer (later), to avoid deadlock with a "bidirectional channel"

                if (ready_to_send_late) { // Send now if ready

                    if_port_commands.write_port (PORT_1D_1N, port_low);
                    ready_to_send_late = false;

                    // Send: trio (to be received on ch_yx in receive_twice_in_y)

                    send_twice_in_x (
                            ch_yx,
                            mic_samples_info  [iof_buffer_to_send_late],
                            buffer            [iof_buffer_to_send_late]);

                    get_next_mic_frame = false;

                    mic_samples_info_init (mic_samples_info[iof_buffer_to_send_late]);

                } else {
                    // get_next_mic_frame is true. No code, send when it's ready
                }

            } break;

            case ch_in :> phase1_sch_mic_sample : { // Match Handle_Beep_BRRRR_task_b
                bool became_full_now;

                // ===================================================================================
                // Observe that there is no PHASE1 -> PHASE2 -> PHASE3 here!
                // Either there are
                //     NUM_000:           N * phase1_sch_mic_sample or
                //     NUM_001 or larger: 1 * phase1_sch_mic_sample then 1 * phase3_ch_mic_samples
                // ===================================================================================
                if (phase1_sch_mic_sample.chan_mutual_packet_dim == NUM_000) {

                    // Fill data into first position for phase3_ch_mic_samples buffer
                    //                             ####### union, so any would go:
                    phase3_ch_mic_samples.samples.rec_001.mic_samples.samples[0] = phase1_sch_mic_sample.iff.valid.mic_sample;
                    phase3_ch_mic_samples.samples.rec_001.nums.num_samples       = 1;
                    debug_one_sample_cnt++;

                    VERIFY_MIC_SAMPLES_DELTA (10, 1, debug_sliding_mic_sample, phase1_sch_mic_sample.iff.valid.mic_sample, NUM_000);

                    became_full_now =
                        mic_samples_info_fill_buffer (
                            phase3_ch_mic_samples,
                            buffer[iof_buffer_to_fill],
                            mic_samples_info[iof_buffer_to_fill]);

                } else { // >= NUM_001 meaning phase1_sch_mic_sample.iff.valid.is_void = MIC_SAMPLE_VOID;

                    receive_variant_protocol (
                            ch_in,
                            phase1_sch_mic_sample.chan_mutual_packet_dim,
                            phase3_ch_mic_samples); // For mic_samples_info_fill_buffer

                    const unsigned     num_samples = phase3_ch_mic_samples.samples.rec_max_msg.nums.num_samples;
                    const mic_sample_t next_sample = phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[num_samples-1];

                    VERIFY_MIC_SAMPLES_DELTA (12, num_samples, debug_sliding_mic_sample, next_sample, -1);

                    debug_all_samples_cnt++;

                    became_full_now =
                       mic_samples_info_fill_buffer (
                            phase3_ch_mic_samples,
                            buffer[iof_buffer_to_fill],
                            mic_samples_info[iof_buffer_to_fill]);
                }

                if (became_full_now) {
                    if (get_next_mic_frame) {

                        mic_samples_info[iof_buffer_to_fill].was_late = false;

                        // Send [iof_buffer_to_fill] since this was the last filled:
                        // trio (to be received on ch_yx receive_twice_in_y)
                        send_twice_in_x (
                                ch_yx,
                                mic_samples_info  [iof_buffer_to_fill],
                                buffer            [iof_buffer_to_fill]);

                        get_next_mic_frame = false;

                        // Buffer gone, keep filing into iof_buffer_to_fill

                    } else { // Not asked for yet, set buffer aside

                        // This is related to LOGGER_PERIODE_CMD_A_TICKS, which is related to the
                        // speed of how fast the I2C is and how many bytes are written there.
                        // Getting this "late" status here is really no problem, as processing is not related
                        // to the logger. If the logger is too slow, processing is done, but the logger
                        // doesn't get that curve. However, to have the logger do EVERY curve,
                        // LOGGER_PERIODE_CMD_A_TICKS must be correct, and it boils down to the sampling rate
                        // PROCESS_FREQUENCY_HZ and NUM_SAMPLES_PER_BATCH, how many spectrum bins that are
                        // done per FFT, per the time needed to get the logger to process the spectrum
                        // fast enough (BEEP-006)

                        if_port_commands.write_port (PORT_1D_1N, port_high);

                        ready_to_send_late                            = true;
                        mic_samples_info[iof_buffer_to_fill].was_late = true; // in buffer_task_x
                        iof_buffer_to_send_late                       = iof_buffer_to_fill;


                        // For the the next sample we need to change buffer and fill into the other:
                        iof_buffer_to_fill = (iof_buffer_to_fill + 1) % NUM_BUFFERS;
                    }

                    mic_samples_info_init (mic_samples_info[iof_buffer_to_fill]);

                    debug_one_sample_cnt  = 0;
                    debug_all_samples_cnt = 0;
                } else {}

            } break;
        }
    }
} // buffer_task_x


typedef struct {
    bool mul_N; // Handled first
    bool sqrt;  // second
    bool dB;    // third: irregardless of do_sqrt, this may run alone or after do_sqrt
} do_calcs_t;


bool my_dsp_magnitude ( // return ok: no overflow/underflow or Q24_NaN
        const int32_t       mul_N, // signed because of usage
        const unsigned      iof_spectrum, // not used yet
        const dsp_complex_t freq_spectrum_in [NUM_FREQ_POINTS_COMPLEX], // ..same size..
        const do_calcs_t    do_calcs,
        spectrum_val_t      spectrum_val_out [NUM_FREQ_POINTS_REAL], // ..as this
        int_max_value_t    &freq_max_value,
        error_info_t        &error)
{
    error_info_t error_prev;
    error_info_t error_round;

    for (unsigned ix=0; ix<NUM_FREQ_POINTS_COMPLEX; ix++) {
        error_prev = error;
        error_round = error;

        dsp_complex_t val; // comes at no cost (I think), so for readability
        val.re = freq_spectrum_in[ix].re;
        val.im = freq_spectrum_in[ix].im;

        if (do_calcs.mul_N) {
            // If-test her instead of in a separate loop, since I have set
            // freq_spectrum_in as const. I think it's "nice" to not modify it
            //
            // http://www.dspguide.com/CH31.PDF 1/N for both re and im
            //
            val.re = my_math_multiply_32bit_sat (val.re, mul_N, error);
            val.im = my_math_multiply_32bit_sat (val.im, mul_N, error);
        } else {}

        if (error.err_cnt == error_prev.err_cnt) { // OK
            error_prev = error;
            // hypotenuse squared = mag*mag = re*re + im*im
            //
            spectrum_val_out[ix]  = my_dsp_math_multiply_sat (val.re, val.re, SPECTRUM_Q_FORMAT, error);
            spectrum_val_out[ix] += my_dsp_math_multiply_sat (val.im, val.im, SPECTRUM_Q_FORMAT, error);

            if (error.err_cnt == error_prev.err_cnt) { // OK
                error_prev = error;

                if (do_calcs.sqrt) {
                    // Does not update error:
                    spectrum_val_out[ix] = (spectrum_val_t) dsp_math_sqrt ((uq8_24)spectrum_val_out[ix]);
                } else {}

                if (do_calcs.dB) {
                    spectrum_val_out[ix] = my_dB ((uq8_24)spectrum_val_out[ix], error);
                } else {}
            } else {}
        } else {}

        // Find max
        //
        if (error_round.err_cnt == error.err_cnt) { // OK
            if (spectrum_val_out[ix] > freq_max_value.val_at_ix) {
                freq_max_value.val_at_ix = spectrum_val_out[ix];
                freq_max_value.ix = ix;
                freq_max_value.val_new_cnt++;
            } else {
                freq_max_value.val_stable_cnt++;
            }
        } else {
            freq_max_value.val_aux_cnt++;
        }
    }

    return not (error.err_cnt != error_prev.err_cnt);
} // my_dsp_magnitude


void time_reverse_samples (mic_sample_t mic_samples [NUM_MIC_SAMPLES]) {

    unsigned     iof_left  = 0;
    unsigned     iof_right = (NUM_MIC_SAMPLES-1);
    mic_sample_t left      = mic_samples[iof_left];
    mic_sample_t right     = mic_samples[iof_right];

    while (iof_left < (NUM_MIC_SAMPLES/2)) {
        mic_samples[iof_left]  = right;
        mic_samples[iof_right] = left;

        iof_left++;
        iof_right--;

        left  = mic_samples[iof_left];
        right = mic_samples[iof_right];
    }
}

void init_int_max_value (int_max_value_t &freq_max_value) {
    // freq_max_value.curve_is_all_zeros Don't touch
    //
    freq_max_value.val_at_ix      = ZERO_VAL;
    freq_max_value.ix             = 0; // in curve
    freq_max_value.val_new_cnt    = 0;
    freq_max_value.val_stable_cnt = 0;
    freq_max_value.val_aux_cnt    = 0;
} // init_int_max_value

#define NUM_BATCHES_TO_COUNT     50
#define MARGIN_MEAN_BATCH_CNT_MS  0

void init_error (error_info_t &error) {
    error.err_cnt   = 0;
    error.debug_tag = 0;
}


// Some concistency checks

#if ((NUM_COMPLEX_PARTS * NUM_FFT_POINTS_COMPLEX) != NUM_MIC_SAMPLES)
    #error ALL UNIONS MUST BE OF EQUAL SIZE
#elif ((NUM_COMPLEX_PARTS * NUM_FREQ_SPECTRA * NUM_FREQ_POINTS_REAL) != NUM_MIC_SAMPLES)
    #error ALL UNIONS MUST BE OF EQUAL SIZE
#else // This needs to be a thousand percent correct:
    // xassert (sizeof (mic_samples_buff_t) BELOW
#endif

#if (DO_ONE_FRAME_DELAY == 1)
    #define NUM_FRAMES 2 // Largest possible since "new" and "old" are two (coding)
#else
    #define NUM_FRAMES 1
#endif

void init_frames (mic_samples_buff_t frames[NUM_FRAMES]) {
    for (unsigned ix=0; ix<NUM_FRAMES; ix++) {
        for (unsigned iy=0; iy<NUM_MIC_SAMPLES; iy++) {
            frames[ix].u.mic_samples [iy] = Q24_NaN; // 1 at least won't cause div by zero
        }
    }
}

#define DROP_THESE_FRAMES_CNT 3

typedef struct {
    mic_samples_info_t   mic_samples_info  [NUM_FRAMES];
    mic_samples_buff_t   frames            [NUM_FRAMES];
    unsigned             drop_this_frame_cntdown;
    unsigned             iof_frame_newest;
    unsigned             iof_frame_oldest;
    get_next_mic_frame_t get_next_mic_frame;
    bool                 any_button_any_event;
    int_max_value_t      freq_max_value;
    ch_data_ack_ui_t     ch_data_ack_ui;

    #if (DEBUG_MIC_SAMPLES_INCREASING > 0)
        mic_sample_t debug_sliding_mic_sample;
    #endif
    #if (NUM_SECTIONS_FILTER_Y > 0)
        iir_biquad_params_t filter_y_iir_biquad_params;
    #endif
} task_y_context_t;


void GET_NEW_DATA_FROM_X (
        chanend          ch_yx,
        const unsigned   num_frames,
        task_y_context_t &ctx)
{
    ctx.iof_frame_newest = (ctx.iof_frame_newest+1) % num_frames; // 0 if 1 or 1/0 if NUM_FRAMES
    ctx.iof_frame_oldest = (ctx.iof_frame_oldest+1) % num_frames; // 0 if 1 or 0/1 if NUM_FRAMES

    // One query, one answer, to avoid deadlock with a "bidirectional channel"
    // To be received on ch_yx
    ch_yx <: ctx.get_next_mic_frame; // = true. In dsp_task_y sends to buffer_task_x

    // It's ok not to have the first one of the one below in a select since
    // all the other comms with this are in a select here (sch_yb and ch_logger)
    //
    // Receive trio (was sent on ch_yx by send_twice_in_x)
    //
    receive_twice_in_y (
            ch_yx,
            ctx.mic_samples_info [ctx.iof_frame_newest],
            ctx.frames           [ctx.iof_frame_newest]);

    xassert (ctx.mic_samples_info[ctx.iof_frame_newest].num_samples == NUM_MIC_SAMPLES); // 1024 or 2048 signed

    // More debug prints here in v0618. For that version, tested and OK:
    #if (DEBUG_MIC_SAMPLES_INCREASING > 0)
        for (unsigned ix = 0; ix < ctx.mic_samples_info[ctx.iof_frame_newest].num_samples; ix++) {
            VERIFY_MIC_SAMPLES_DELTA (30, 1,
                    ctx.debug_sliding_mic_sample,
                    ctx.frames[ctx.iof_frame_newest].u.mic_samples[ix],
                    ctx.mic_samples_info[ctx.iof_frame_newest].num_samples);
        }
    #endif
}


void HANDLE_PROCESS_FFT_ETC (
        task_y_context_t &ctx)
{
    #if (NUM_SECTIONS_FILTER_Y > 0)
        for (unsigned ix=0; ix < ctx.mic_samples_info[ctx.iof_frame_oldest].num_samples; ix++) {
            // Anti-alias 16 kHz to 4 kHz
            ctx.frames[ctx.iof_frame_oldest].u.mic_samples[ix] =
                    dsp_filters_biquads (
                        ctx.frames[ctx.iof_frame_oldest].u.mic_samples[ix],
                        ctx.filter_y_iir_biquad_params.coeffs,
                        ctx.filter_y_iir_biquad_params.states,
                        NUM_SECTIONS_FILTER_Y,
                        FILTER_Q_FORMAT);
        }
    #endif

    // TODO hanning/hamming window here?
    // Windowing here, before bit reversing. See lib_mic_array user guide page 17.

    // frame is a union:
    // mic_sample_t  mic_samples          [NUM_MIC_SAMPLES];                        // 1 *     512, 1024 or 2048
    // dsp_complex_t complex_array        [NUM_FFT_POINTS_COMPLEX];                 // 2 *     256,  512 or 1024
    // dsp_complex_t complex_half_spectra [NUM_FREQ_SPECTRA][NUM_FREQ_POINTS_REAL]; // 2 * 2 * 128,  256 or  512

    // ==== From dsp_lib ====
    dsp_fft_bit_reverse (
            ctx.frames[ctx.iof_frame_oldest].u.complex_array,  // 512, 1024 or 2048 signed seen as 512 or 1024 complex
            NUM_FFT_POINTS_COMPLEX); // 256,  512 or 1024

    // ==== From dsp_lib ====
    dsp_fft_forward (
            ctx.frames[ctx.iof_frame_oldest].u.complex_array,
            NUM_FFT_POINTS_COMPLEX,            // 256, 512 or 1024 complex
            FFT_SINE(NUM_FFT_POINTS_COMPLEX)); // dsp_sine_256  is dim ( 256/4) + 1 =  65
                                               // dsp_sine_512  is dim ( 512/4) + 1 = 129
                                               // dsp_sine_1024 is dim (1024/4) + 1 = 257

    // NUM_FFT_POINTS_COMPLEX fed into my_dsp_magnitude below
    // The magnitude of the FFT output is right shifted log2(N) times which corresponds to
    // division by N as shown in EQUATION 31-5 of http://www.dspguide.com/CH31.PDF.

    // ==== From dsp_lib ====
    dsp_fft_split_spectrum (
            ctx.frames[ctx.iof_frame_oldest].u.complex_array,
            NUM_FFT_POINTS_COMPLEX);
}


void HANDLE_PROCESS_SPECTRUM (
        task_y_context_t          &ctx,
        const unsigned            iof_spectrum,
        ch_data_logger_spectrum_t &ch_data_logger_spectrum)
{

     bool             ok;
     do_calcs_t       do_calcs;
     error_info_t     error;

     do_calcs.mul_N = false;
     do_calcs.sqrt  = true; // A ton of scaling problems if this is not set!
     do_calcs.dB    = false;

     init_error (error);
     init_int_max_value (ctx.freq_max_value);

     ok = my_dsp_magnitude (
             NUM_FFT_POINTS_COMPLEX, // if do_calcs.mul_N true
             iof_spectrum,
             ctx.frames[ctx.iof_frame_oldest].u.complex_half_spectra[iof_spectrum],
             do_calcs,
             ch_data_logger_spectrum.spectrum,
             ctx.freq_max_value,
             error);

    #if (DEBUG_DSP_ERROR_LOG==1)
        if (not ok) {
            debug_print ("tag1[%X], err:%u\n",
                    error.debug_tag,
                    error.err_cnt);
        }
    #endif
}

void HANDLE_SEND_TO_LOGGER (
        chanend                   ch_logger,
        task_y_context_t          &ctx,
        const unsigned            iof_spectrum,
        ch_data_logger_spectrum_t &ch_data_logger_spectrum)
{
     ch_data_logger_spectrum.iof_spectrum   = iof_spectrum;
     ch_data_logger_spectrum.freq_max_value = ctx.freq_max_value;
     //
     ch_logger <: ch_data_logger_spectrum;
}

void HANDLE_KNOCK_TO_TASK_B (
        streaming chanend sch_yb,
        const             knock_reason_e knock_reason)
{
    data_sch_knock_t data_sch_knock;
    data_sch_knock.knock_reason = knock_reason;

    // Atomic deadlock-free exchange of data
    sch_yb <: data_sch_knock; // Will never block, ie. no deadlock
}

void HANDLE_UI_DATA_FROM_TASK_B (
        task_y_context_t &ctx) // ctx.ch_data_ack_ui is NEW data
{
    // The ONLY UI data at the moment is whether a button has been pressed!

    if (ctx.ch_data_ack_ui.any_button_any_event.cnt != ctx.any_button_any_event) {
        ctx.any_button_any_event = ctx.ch_data_ack_ui.any_button_any_event.cnt;
        //
        // Some button pressed or released, if pressed for long assuming silence
        //
        ctx.drop_this_frame_cntdown = DROP_THESE_FRAMES_CNT;
    } else {}

    ctx.freq_max_value.curve_is_all_zeros = (ctx.drop_this_frame_cntdown > 0);

    if (ctx.drop_this_frame_cntdown > 0) {
        /* TODO remove
        debug_print_special ("%s%u%s",
                (ctx.drop_this_frame_cntdown == DROP_THESE_FRAMES_CNT) ? "z " : "",
                 ctx.drop_this_frame_cntdown,
                (ctx.drop_this_frame_cntdown == 1) ? "\n" : " ");
        */
        ctx.drop_this_frame_cntdown--;
    } else {}
}

void HANDLE_SEND_RESULT_TO_TASK_B (
        streaming chanend     sch_yb,
        const int_max_value_t freq_max_value_result)
{
    sch_yb <: freq_max_value_result; // My data from FFT etc.
}


typedef struct {
    time32_t for_mean_time_ticks;
    time32_t time_start_dsp_ticks;
    time32_t time_end_dsp_ticks;
    unsigned dsp_time_sum_us;
    unsigned batch_cnt;
    unsigned was_late_cnt;
} timing_t;


void HANDLE_ANALYZE_TIMING (
        timing_t         &timing,
        task_y_context_t &ctx) {

    timer tmr;

    timing.dsp_time_sum_us += ((timing.time_end_dsp_ticks - timing.time_start_dsp_ticks) / XS1_TIMER_MHZ);

    timing.batch_cnt++;

    if (ctx.mic_samples_info[ctx.iof_frame_oldest].was_late) { // In dsp_task_y
        timing.was_late_cnt++;
    } else {}

    if ((timing.batch_cnt % NUM_BATCHES_TO_COUNT) == 0) {
        time32_t for_mean_time_ticks_now;

        tmr :> for_mean_time_ticks_now;

        const unsigned mean_batch_time_ms = ((for_mean_time_ticks_now - timing.for_mean_time_ticks)/XS1_TIMER_KHZ) / NUM_BATCHES_TO_COUNT;

        #if (DEBUG_DSP_STATUS_LOG == 1)
            // char magnitude_str[Q_FORMAT_STR_LEN];
            // q24_signed_to_str (magnitude_str, ctx.freq_max_value.val_at_ix);

            // Halts processor every 6.5 secs
            // in dsp_task_y
            /*                 "late=" (the other is "late:")
            debug_print ("%d ms late=%u dsp:%u us max:%s %d(%#08x) but:%u\n",
                    mean_batch_time_ms,
                    timing.was_late_cnt,
                    dsp_time_sum_us/NUM_BATCHES_TO_COUNT,
                    magnitude_str, ctx.freq_max_value.val_at_ix, ctx.freq_max_value.val_at_ix,
                    ctx.ch_data_ack_ui[ctx.iof_frame_newest].any_button_any_event.cnt);
            */
            //                 "late=" (the other is "late:")
            debug_print ("%d ms late=%u\n",
                   mean_batch_time_ms,
                   timing.was_late_cnt);
        #endif

        timing.was_late_cnt = 0;
        timing.dsp_time_sum_us = 0;

        timing.for_mean_time_ticks = for_mean_time_ticks_now;
    }
}


// dsp_state transition diagram
//
// get_new_data_from_x:     dsp_state = process_fft_etc;
// process_fft_etc:         dsp_state = loop_get_ui_data_from_b;
// loop_get_ui_data_from_b: dsp_state = loop_process_spectrum;
// loop_process_spectrum:   dsp_state = loop_send_to_logger;
// loop_send_to_logger:     dsp_state = loop_send_result_to_b;
// loop_send_result_to_b:   dsp_state = loop_until_analyze;
// loop_until_analyze:      dsp_state = loop_get_ui_data_from_b;
//                          dsp_state = get_new_data_from_x;
//
typedef enum {
    get_new_data_from_x,
    process_fft_etc,
                           // Loop NUM_FRAMES times
    loop_get_ui_data_from_b,
    loop_process_spectrum,
    loop_send_to_logger,
    loop_send_result_to_b,
    loop_until_analyze     // -> get_new_data_from_x when NUM_FRAMES done
} dsp_state_e;


void HANDLE_RESPONSE_FROM_TASK_B (
        streaming chanend      sch_yb,
        const task_y_context_t ctx,
        const dsp_state_e      dsp_state,
        const int_max_value_t  freq_max_value_result)
{
    switch (ctx.ch_data_ack_ui.data_sch_ack.ack_reason) {
        case in_ui: { // after in_ui_after_ack
            xassert (dsp_state == loop_process_spectrum);
        } break;
        case out_dsp_result: { // after out_dsp_result_after_ack
            xassert (dsp_state == loop_until_analyze);
            //
            HANDLE_SEND_RESULT_TO_TASK_B (sch_yb, freq_max_value_result);
        }  break;
        default: xassert (false); break;
    }
}

// [[combinable]] // Not with ordered select
void dsp_task_y (
        chanend           ch_yx,
        chanend           ch_logger,
        streaming chanend sch_yb)
{
    xassert (sizeof (data_sch_knock_t)   <= MAX_SIZEOF_NONBLOCKING_STREAMING_CHAN);
    xassert (sizeof (mic_samples_buff_t) == (sizeof (int64_t) + (NUM_MIC_SAMPLES * (sizeof (int32_t)))));

    debug_print ("dsp_task_y started, sizeof mic_samples_buff_t: %u\n", (sizeof (mic_samples_buff_t)));

    task_y_context_t ctx; // Values that may change in the HANDLE_ functions

    ch_data_logger_spectrum_t ch_data_logger_spectrum;
    int_max_value_t           freq_max_value_result;

    unsigned    iof_spectrum    = 0;
    bool        wait_for_task_b = false;
    dsp_state_e dsp_state       = get_new_data_from_x;
    unsigned    num_frames      = NUM_FRAMES-1;

    next_curve_logger_state_e next_curve_logger_state = logger_never_seen;

    ctx.iof_frame_newest        = 0; // equal on first round, effectively..
    ctx.iof_frame_oldest        = 0; // .. one frame only, then num_frames changes to NUM_FRAMES
    ctx.get_next_mic_frame      = true;
    ctx.any_button_any_event    = 0;
    ctx.drop_this_frame_cntdown = 0;

    ctx.freq_max_value.curve_is_all_zeros = false;
    init_int_max_value (ctx.freq_max_value);

    init_frames (ctx.frames);

    #if (NUM_SECTIONS_FILTER_Y > 0)
        // Init filter
        dsp_design_biquad_lowpass (
                FILTER_Y_FREQ_MINUS_3DB_NORMALIZED,
                FILTER_Y_Q_FACTOR,
                ctx.filter_y_iir_biquad_params.coeffs,
                FILTER_Q_FORMAT);
        //
        // Initialize filter state data to zeros
        //
        for (int32_t ix = 0; ix < DSP_NUM_STATES_PER_BIQUAD; ++ix) {
            ctx.filter_y_iir_biquad_params.states[ix] = 0;
        }
    #endif

    timer    tmr;
    time32_t time_ticks;

    timing_t timing;
    //
    timing.dsp_time_sum_us = 0;
    timing.batch_cnt       = 0;
    timing.was_late_cnt    = 0;

    tmr :> timing.for_mean_time_ticks;
    tmr :> time_ticks;

    while (1) {
        [[ordered]]
        select {
            case ch_logger :> next_curve_logger_state : { // == logger_next_curve_request
                // logger_task finished with previous curve "already"
                // If I2C failed then this message is not received, ie. logger_never_seen stays
                //
                tmr :> time_ticks; // Ready immediately
            } break;

            case sch_yb :> ctx.ch_data_ack_ui : {
                HANDLE_RESPONSE_FROM_TASK_B (sch_yb, ctx, dsp_state, freq_max_value_result);
                wait_for_task_b = false;
                tmr :> time_ticks; // Ready immediately
            } break;

            case (not wait_for_task_b) => tmr when timerafter (time_ticks) :> void : {
                debug_print ("%ss: %u, i:%u\n", // See log "2022 06 20 v0834 log.txt"
                        (dsp_state == get_new_data_from_x) ? "\n" : "",
                        dsp_state,
                        iof_spectrum);

                switch (dsp_state) {
                    case get_new_data_from_x: {
                        GET_NEW_DATA_FROM_X (ch_yx, num_frames, ctx);
                        dsp_state = process_fft_etc;
                    } break;

                    case process_fft_etc: {
                        tmr :> timing.time_start_dsp_ticks;
                        HANDLE_PROCESS_FFT_ETC (ctx);
                        tmr :> timing.time_end_dsp_ticks;
                        dsp_state = loop_get_ui_data_from_b;
                    } break;

                    case loop_get_ui_data_from_b: {
                        HANDLE_KNOCK_TO_TASK_B (sch_yb, in_ui_in_ack);
                        wait_for_task_b = true;
                        dsp_state = loop_process_spectrum;
                    } break;

                    case loop_process_spectrum: {
                        HANDLE_UI_DATA_FROM_TASK_B (ctx); // ctx.ch_data_ack_ui is NEW data
                        HANDLE_PROCESS_SPECTRUM (ctx, iof_spectrum, ch_data_logger_spectrum);
                        dsp_state = loop_send_to_logger;
                    } break;

                    case loop_send_to_logger: {
                        if (next_curve_logger_state == logger_never_seen) {
                            // No code
                        } else {
                            if (next_curve_logger_state == logger_next_curve_no_request) {
                                // If logger_task not finished with previous curve, wait for it here
                                ch_logger :> next_curve_logger_state; // Will be logger_next_curve_request
                            } else {} // No code

                            if (next_curve_logger_state == logger_next_curve_request) {
                                HANDLE_SEND_TO_LOGGER (ch_logger, ctx, iof_spectrum, ch_data_logger_spectrum);
                            } else {}
                            next_curve_logger_state = logger_next_curve_no_request;
                        }
                        dsp_state = loop_send_result_to_b;
                    } break;

                    case loop_send_result_to_b: {
                        freq_max_value_result = ctx.freq_max_value;
                        //
                        HANDLE_KNOCK_TO_TASK_B (sch_yb, out_dsp_result_after_ack);
                        wait_for_task_b = true;
                        dsp_state = loop_until_analyze;
                    } break;

                    case loop_until_analyze: {
                        if (iof_spectrum < (NUM_FREQ_SPECTRA - 1)) { // < 1 is 0!
                            // DO THE SECOND CURVE AS WELL:
                            iof_spectrum++; // Essentially from 0 to 1
                            dsp_state = loop_get_ui_data_from_b;
                        } else { // == 1
                            // LAST CURVE OF THE TWO:
                            iof_spectrum = 0;
                            HANDLE_ANALYZE_TIMING (timing, ctx);
                            dsp_state = get_new_data_from_x;
                        }
                        // Start all next rounds after the first state_process_spectrum NUM_FRAMES was 1.
                        // After this ctx.iof_frame_newest and ctx.iof_frame_oldest will be skewed by (NUM_FRAMES-1)
                        //
                        num_frames = NUM_FRAMES; // become a var constant. Never modified after this
                    } break;

                    default: xassert (false); break;
                } // switch dsp_state

                tmr :> time_ticks; // Ready immediately
            }  break; // select case
        } // select
    }
} // dsp_task_y
