/*
 * logger.xc
 *
 *  Created on: 11. feb. 2022
 *      Author: oyvindteig
 */

#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <xscope.h>   // xScope logging
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <i2c.h>
    #include <i2s.h>
    #include <dsp_qformat.h>
    #include <dsp.h>
    #include <dsp_fft.h>
    #include <dsp_math.h>
    #include <math.h> // log10 etc
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "maths.h"
    #include "mic_array.h"
    #include "maths_fix.h" // Needs stdint.h
    #include "mics_in_headset_out.h"
    #include "param.h"
    #include "i2c_client_task.h"
    #include "DAC4_mcp4728.h"
    #include "port_tile0_task.h"
    #include "_Beep_dsp_handling.h"
    #include "logger.h"

    #define DO_DEBUG_PRINT         0 // 0 has consequences if (USE_XSCOPE == 1) and amy timing at all needs to be met
    #define DO_DEBUG_PRINT_SPECIAL 0 // 1: debug_print_special
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif

#if (WARNINGS==1)
    #ifndef USE_XSCOPE
        #error USE_XSCOPE not defined in makefile
    #elif (USE_XSCOPE == 0)
        #warning NO XSCOPE
    #elif ((USE_XSCOPE==1) and (USE_XSCOPE_AT==XSCOPE_AT_2))
        #warning USING XSCOPE AT "STARTSTOP UINT"
    #endif
#endif

// ======================================================================
// ABOUT XSCOPE
// ======================================================================
// xScope timing is for a fill width Window (with 10 parts):
//
// us:  1, 2, 5, 10, 20, 50, 100, 200, 500
// ms:  --"--
// sec: 1, 2, 5, 10, 20, 50
//
// Also observe file "config.xscope" and the xTIMEcomposer User Guide

// xSCOPE beware of OFFSET which for the two first is just a number:
// -----------------------------------------------------------------
// Samples/div |                                    OFFSET
// Always      | Rising   | Falling |               OFFSET
// Auto        | Separate | Persist | Save to png
// Window      | Diff     |                         OFFSET
// -----------------------------------------------------------------
//

// In the xTIME composer User Guide chapter "19xSCOPE performance figures"
// we see that max some 500-900 values per second are picked up by xScope.
// Even if I use XTAG-3 I use an old machine (Mac Mini from 2010)

// So, testing max 500 per second, or 256 inside 500 ms window
// SEEMS LIKE I NEED TO HAVE THE WINDOW ON 1 SEC?


// ======================================================================
// ABOUT I2C
// ======================================================================
// Observe that we have gone to great lengths to store MIC data while the i2C
// is busy to write to the display, see MAX_WAIT_TIME_TO_SEND_DATA_MS. So we do have
// to introduce some load balancing. I'd like the logging_task below to be linear,
// and I'd like it not interfering with mics_in_headset_out_task_a too much by it
// competing with Display_screen

#ifndef I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND
    #error missing
#else
    #define LOGGER_PERIODE_CMD_A_TICKS_ ((XS1_TIMER_MHZ * DAC4_REG_FAST_WRITE_CMD_A_NUM_BITS * 1000) / I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND)
            // 100 * 29*1000/400 = 100 * 72.5 -> 72 us (400 kHz I2C)
    #if        (LOGGER_PERIODE_CMD_A_TICKS_ < (XS1_TIMER_MHZ * 73))
        #define LOGGER_PERIODE_CMD_A_TICKS    (XS1_TIMER_MHZ * 100) // 100 us
        //                                                     100 See BEEP-006 for a discussion, however:
        //                                                     100 fast enough to do every curve
        //                                                     100 long enough to have time for the I2C needed
        // PROCESS_FREQUENCY_HZ   16000
        // NUM_SAMPLES_PER_BATCH   1024 is 64   ms per batch
        // NUM_FREQ_POINTS_REAL     256 is 32   ms per spectrum up to 4000 Hz
        // LOGGER:         100 us * 256 is 25.6 ms per spectrum

    #else
        #warning Find another value!
        #define LOGGER_PERIODE_CMD_A_TICKS (XS1_TIMER_MHZ * 100)
    #endif

    #define XSCOPE_PERIODE_TICKS (XS1_TIMER_MHZ * 1900) // 1.9 ms for NUM_FREQ_POINTS_REAL (256) all in 486 ms
#endif



[[combinable]]
void logger_task (
        chanend                        ch_logger, // request_next_curve_to_logger out, curve in
        client i2c_general_commands_if if_i2c_general_commands,
        client port_commands_if        if_port_commands)
{
    debug_print ("logger_task started I2C %u\n", LOGGER_TASK_ALLOW_I2C);

    timer    tmr;
    time32_t time_ticks;
    time32_t timeerafter_ticks;

    const next_curve_logger_state_e next_curve_logger_state = logger_next_curve_request;

    ch_data_logger_spectrum_t ch_data;;

    DAC4_context_t DAC4_ctx;
    i2c_uint8_t    i2c_bytes [MUM_DAC4_ABCD_CHANNELS * 2];
    bool           i2c_error;
    unsigned       iof_value = NUM_LOGGER_TASK_CURVE_VALUES; // 0;
    bool           do_curve = false;
    bool           debug_zero_val_pulse_port_cnt = 0;
    bool           trigger_port = false;

    const unsigned do_scale_ratio_all [MUM_DAC4_ABCD_CHANNELS] = {true, true, true, true};
    //
    for (unsigned ix=DAC4_CH0_A; ix<MUM_DAC4_ABCD_CHANNELS; ix++) {
        DAC4_ctx.do_scale_ratio[ix] = do_scale_ratio_all[ix];
    }

    unsigned  debug_in_time_cnt = 0;
    unsigned  debug_late_cnt = 0;
    unsigned  debug_ok_cnt = 0;
    unsigned  debug_err_cnt = 0;
    unsigned  debug_longest_interval_deltatime_ticks = 0;

    #if (DEBUG_LOGGER_SKIP_CURVES==0) // Standard
        int32_t  freq_max_val_at_ix_max = FREQ_MAX_VALUE_LOG;
        unsigned debug_curves_cnt = 0;
        unsigned debug_ERR_cnt = 0;
        unsigned debug_OK_cnt = 0;
    #endif

    #if (LOGGER_TASK_ALLOW_I2C==1)
        // Test if DAC4 is present
        // Not sending request_next_curve_to_logger true makes up for no logging ever
        // (or sending it false)
        // since it's defaulted to false in dsp_task_y

        DAC4_ctx.curve[DAC4_CH0_A] = ZERO_VAL;
        I2C_DAC4_fill_buffer (DAC4_ctx, i2c_bytes, MUM_DAC4_A_CHANNELS);

        i2c_error = not if_i2c_general_commands.write_reg_ok (
                I2C_ADDRESS_OF_DAC4,
                i2c_bytes,
                MUM_DAC4_A_BYTES);
        if (not i2c_error) {
            ch_logger <: next_curve_logger_state; // ie. logger_next_curve_request
        } else {}
    #else
        i2c_error = false;
    #endif

    while (1) {
        select {
            case (do_curve) => tmr when timerafter (time_ticks) :> timeerafter_ticks : {

                if (iof_value == NUM_LOGGER_TASK_CURVE_VALUES) { // == 256

                    debug_print ("logger curve done%s", "\n");
                    do_curve = false;

                    // So that the last value will not continue until the next spectrum as a "DC value"
                    DAC4_ctx.curve [DAC4_CH0_A] = 0;
                    DAC4_ctx.curve [DAC4_CH1_B] = 0;
                    DAC4_ctx.curve [DAC4_CH3_D] = SCOPE_TRIGGER_OFF_VAL;
                    //
                    I2C_DAC4_fill_buffer (DAC4_ctx, i2c_bytes, MUM_DAC4_ABCD_CHANNELS);

                    #if (LOGGER_TASK_ALLOW_I2C==1)
                        i2c_error = not if_i2c_general_commands.write_reg_ok (
                                I2C_ADDRESS_OF_DAC4,
                                i2c_bytes,
                                MUM_DAC4_ABCD_BYTES);
                    #else
                        i2c_error = false;
                    #endif

                    if (i2c_error) {
                        debug_err_cnt++;
                        debug_print ("no last mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    } else {
                        debug_ok_cnt++;
                        debug_print ("ok last mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    }

                    DO_XSCOPE_STOP (VALUE, XSCOPE_AT_2);

                    ch_logger <: next_curve_logger_state; // ie. logger_next_curve_request
                } else {
                    // See "timerafter to now and skew" in blog note 165

                    if (ch_data.spectrum[iof_value] > ch_data.freq_max_value.val_at_ix) {
                        ch_data.spectrum[iof_value] = ch_data.freq_max_value.val_at_ix;
                    } else {}

                    xassert (ch_data.spectrum[iof_value] <= ch_data.freq_max_value.val_at_ix);

                    if (DAC4_ctx.curve_is_all_zeros[DAC4_CH0_A]) {
                        DAC4_ctx.curve[DAC4_CH0_A] = ZERO_VAL;

                        // For scope if DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE 1
                        // TODO remove
                        trigger_port = not trigger_port;
                        if_port_commands.write_port (PORT_1I_1M, trigger_port);
                        trigger_port = not trigger_port;
                        if_port_commands.write_port (PORT_1I_1M, trigger_port);

                        // TODO for v0831: remove:
                        debug_zero_val_pulse_port_cnt++; // Only BIT0 output to port:

                        #if (DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE_ZERO_ONLY == 1)
                            // For scope if DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE 1
                            if_port_commands.write_port (PORT_1I_1M, debug_zero_val_pulse_port_cnt);
                        #endif
                    } else {
                        DAC4_ctx.curve[DAC4_CH0_A] = ch_data.spectrum[iof_value];
                    }
                    I2C_DAC4_fill_buffer (DAC4_ctx, i2c_bytes, MUM_DAC4_A_CHANNELS);

                    if (iof_value == 0) { // Beginning of curve
                        trigger_port = (ch_data.iof_spectrum == 0);
                        //       |--------------|
                        //       | 0            | 1            | ch_data.iof_spectrum
                        // ------|              |--------------|

                        #if (DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE_ZERO_ONLY == 0) // default
                            // For scope if DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE 1
                            if_port_commands.write_port (PORT_1I_1M, trigger_port);
                        #endif
                    } else {}


                    #if (LOGGER_TASK_ALLOW_I2C==1)
                        i2c_error = not if_i2c_general_commands.write_reg_ok (
                                I2C_ADDRESS_OF_DAC4,
                                i2c_bytes,
                                MUM_DAC4_A_BYTES);  // LOGGER_PERIODE_CMD_A_TICKS = 90  needed for "late:0"
                             // MUM_DAC4_AB_BYTES); // LOGGER_PERIODE_CMD_A_TICKS = 130 needed for "late:0"
                             // Using only DAC4_A anyhow! Using 100 us solved BEEP-006 (v0828)
                    #else
                        i2c_error = false;
                    #endif

                    if (i2c_error) {
                        debug_err_cnt++;
                        debug_print ("no mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    } else {
                        debug_ok_cnt++;
                        debug_print ("ok mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    }

                    DO_XSCOPE_INT (VALUE, ch_data.spectrum[iof_value], XSCOPE_AT_2);

                    // This may have taken too long. Don't let time_ticks_now suffer:
                    //
                    time32_t time_ticks_now;
                    tmr :> time_ticks_now;

                    const unsigned used_deltatime_ticks = (time_ticks_now - timeerafter_ticks);
                    //
                    if (used_deltatime_ticks > debug_longest_interval_deltatime_ticks) {
                        debug_longest_interval_deltatime_ticks = used_deltatime_ticks;
                    } else {}

                    if (used_deltatime_ticks < LOGGER_PERIODE_CMD_A_TICKS) {
                        time_ticks += LOGGER_PERIODE_CMD_A_TICKS; // Advance linearly
                        debug_in_time_cnt++;
                    } else {
                        // I call this BEEP-008:
                        // This might happen when BUTTON_ACTION_PRESSED and the Display_screen
                        // is written to. As long as we the DAC4 and the display is sharing the same
                        // I2C lines, there s nothing to do to avoid this. Using an [[ordered]] select
                        // in I2C_Client_Task would not solve the general case, since DAC4 might
                        // to output while Display_screen is actively using it.
                        //
                        // With I2C congestion the value here is some 480 us when the
                        // LOGGER_PERIODE_CMD_A_TICKS is 100 us (v0830)
                        //
                        // TODO: Invent some way to tell that this is not a correct lengthed spectrum value
                        //       But then, the total length is longer
                        //
                        time_ticks = time_ticks_now; // reset time, causes unavoidable nonlinearity
                        debug_late_cnt++;
                    }

                    iof_value++; // To NUM_LOGGER_TASK_CURVE_VALUES which is past index and caught above
                }
            } break;

            case ch_logger :> ch_data : { // Only if request_next_curve_to_logger true sent

                #if (DEBUG_LOGGER_SKIP_CURVES==1) // BEEP-006 and BEEP-007 testing
                    trigger_port = (ch_data.iof_spectrum == 0);
                    //       |--------------|
                    //       | 0            | 1            | ch_data.iof_spectrum
                    // ------|              |--------------|
                    #if (DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE_ZERO_ONLY == 0) // default
                        // For scope if DEBUG_SPECTRUM_VIA_DAC4_ON_SCOPE 1
                        if_port_commands.write_port (PORT_1I_1M, trigger_port);
                    #endif
                    ch_logger <: request_next_curve_to_logger;
                #else

                    const bool all_zeros = (ch_data.freq_max_value.curve_is_all_zeros);

                    {
                        #include "_global_bug_fix.h"
                        // BEEP-009: no diagnostic value, won't hurt and will never happen:
                        DUMMY_XASSERT_183751_V0838_LOGGER_TASK;
                    }

                    ch_data.freq_max_value.val_at_ix = freq_max_val_at_ix_max; // ..then this: Overwrite it
                    // Last with auto-scaling was v0813

                    debug_curves_cnt++;
                    if ((debug_curves_cnt % 1000) == 0) {
                        #if (LOGGER_TASK_ALLOW_I2C==1)  //        "late:" (the other is "late=")
                            debug_print_special ("per_us:%u tim:%u late:%u  OK:%u ERR:%u  ok:%u err:%u max_us:%u, max:%u\n",
                                    LOGGER_PERIODE_CMD_A_TICKS/XS1_TIMER_MHZ,
                                    debug_in_time_cnt, debug_late_cnt,
                                    debug_OK_cnt,      debug_ERR_cnt,
                                    debug_ok_cnt,      debug_err_cnt,
                                    debug_longest_interval_deltatime_ticks/XS1_TIMER_MHZ,
                                    freq_max_val_at_ix_max);
                        #endif

                        debug_in_time_cnt = 0;
                        debug_late_cnt    = 0;
                        debug_OK_cnt      = 0;
                        debug_ERR_cnt     = 0;
                        debug_ok_cnt      = 0;
                        debug_err_cnt     = 0;
                        //
                        debug_longest_interval_deltatime_ticks = 0;
                    } else {
                        debug_print ("logger curve new %u\n", debug_curves_cnt);
                    }

                    unsigned scale_shift_right_by_bits =  find_downscale_by_bits_param (
                            MAX_Q8_24_IF_POSITIVE_NUM_BITS, // 31
                            DAC4_NUM_BITS,                  // 12
                            ch_data.freq_max_value.val_at_ix);      // even if signed q8_24 it's never negative

                    // This basically finds the scaling factor(s) for the curve (to be used above),
                    // plus sets up a framing full widt "pulse" at DAC4_CH3_D
                    // DAC4_CH0_A and DAC4_CH1_B same scaling is ok, since only one is updated with curve values in the loop
                    //
                    const bool     curve_is_all_zeros [MUM_DAC4_ABCD_CHANNELS] = {all_zeros,                        false,                            false, false};
                    const unsigned curve_vals         [MUM_DAC4_ABCD_CHANNELS] = {0,                                0,                                0,     SCOPE_TRIGGER_ON_VAL};
                    const uint32_t numerator_vals     [MUM_DAC4_ABCD_CHANNELS] = {1<<DAC4_NUM_BITS,                 1<<DAC4_NUM_BITS,                 1,     1};
                    const uint32_t denominator_vals   [MUM_DAC4_ABCD_CHANNELS] = {ch_data.freq_max_value.val_at_ix, ch_data.freq_max_value.val_at_ix, 1,     1}; // Arithmetic exception if zero, of course
                    const unsigned shift_bits         [MUM_DAC4_ABCD_CHANNELS] = {scale_shift_right_by_bits,        scale_shift_right_by_bits,        0,     0};
                    //
                    for (unsigned ix=DAC4_CH0_A; ix<MUM_DAC4_ABCD_CHANNELS; ix++) {
                        DAC4_ctx.curve_is_all_zeros        [ix]     = curve_is_all_zeros [ix];
                        DAC4_ctx.curve                     [ix]     = curve_vals         [ix];
                        // DAC4_ctx.do_scale_ratio         [ix]     = do_scale_ratio_all [ix]; done
                        DAC4_ctx.scale_ratio               [ix].num = numerator_vals     [ix];
                        DAC4_ctx.scale_ratio               [ix].den = denominator_vals   [ix];
                        DAC4_ctx.scale_shift_right_by_bits [ix]     = shift_bits         [ix];
                    }

                    xassert (ch_data.freq_max_value.val_at_ix > 0); // To avoid crash quite soon

                    I2C_DAC4_fill_buffer (DAC4_ctx, i2c_bytes, MUM_DAC4_ABCD_CHANNELS);

                    #if (LOGGER_TASK_ALLOW_I2C==1)
                        i2c_error = not if_i2c_general_commands.write_reg_ok (
                                I2C_ADDRESS_OF_DAC4,
                                i2c_bytes,
                                MUM_DAC4_ABCD_BYTES);
                    #else
                        i2c_error = false;
                    #endif

                    if (i2c_error) {
                        debug_ERR_cnt++;
                        debug_print ("NO mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    } else {
                        debug_OK_cnt++;
                        debug_print ("OK mcp4728 %x=%x.%x %x.%x %x.%x %x.%x\n",
                                I2C_ADDRESS_OF_DAC4 << I2C_SHIFT_BITS,
                                i2c_bytes[0],i2c_bytes[1],
                                i2c_bytes[2],i2c_bytes[3],
                                i2c_bytes[4],i2c_bytes[5],
                                i2c_bytes[6],i2c_bytes[7]);
                    }

                    DO_XSCOPE_START (VALUE, XSCOPE_AT_2);
                    do_curve = true;
                    iof_value = 0;
                    tmr :> time_ticks; // reset time
                    time_ticks += LOGGER_PERIODE_CMD_A_TICKS;
                #endif
            } break;
        }
    }
}

