/*
 * maths_fix.h
 *
 * OBSERVE that lib_dsp dsp_math.h + dsp_math.xc have all that's needed here I assume
 *
 *  Created on: 5. nov. 2021
 *      Author: oyvindteig
 */

/*
From lib_dsp.pdf:
    The xCORE-200 has a single cycle 32x32->64 bit multiply/accumulate unit,
    single cycle double-word load and store, dual issue instruction execution,
    and other instruction set enhancements. These features make xCORE-200 an
    efficient platform for executing digital signal processing algorithms.

However, division/divide do have instructions, but they are not single cycle.
See lib_dsp dsp_math_divide which uses an instruction for long unsigned division

I have not found a list of instructions and cycle count but here is some more info:

Assembly Programming Manual
Publication Date: 2015/3/27 Document Number: X9432B

divs  d, x, y          Signed division
divu  d, x, y          Unsigned division
ldivu d, e, v, x, y    Long unsigned divide
lmul  d, e, x, y, v, w Long multiply
maccs d, e, x, y       Mulitply and accumulate signed
maccu d, e, x, y       Multiply and accumulate unsigned
mul   d, x, y          Multiply

Cycle count: https://www.xcore.com/viewtopic.php?f=3&t=2067
64 bit math operations? Clock cycles per operation?
    From 2013 before XS2
    You might be able to look at the disassembly of the code to see how many instructions that line translates in to,
    but if you want to know how fast it with optimizations on you can create a timer in xc.
    They count by default at 100MHz so if you create two variables and grab the timers value before and after the
    multiply operation and subtract the two you get how many cycles it tock to get there.(Might be a good idea to
    copy paste 10 multiplys and measure that so you get a more accurate reading)

Also here: https://www.xcore.com/viewtopic.php?t=5775
LDIVU questions
    And yes, as I understand it, there is only one divide unit per tile, so everything on that tile has to queue up to use it.
    When a division has to be done, I assume 32 * n worst case cycles, n being the number of threads which might use the divider.

Also https://www.xcore.com/viewtopic.php?f=26&t=2475
How do I target the LDIV instruction?
    LDIVU instruction only works if the division fits in a 32-bit word, that is, if the higher word of the double word input
    is less than the divisor. This operation is intended to be used for the implementation of long division.
*/

#ifndef MATHS_FIX_H_
    #define MATHS_FIX_H_

    // Fixed point         https://en.wikipedia.org/wiki/Fixed-point_arithmetic
    // See                 https://schaumont.dyn.wpi.edu/ece4703b20/lecture6.html
    // "fix16" naming from https://en.wikipedia.org/wiki/Libfixmath
    //

    // ========================================================================================
    // 32 bits = 16 bits POS and 16 bits NEG:Should make sense if MIC_ARRAY_WORD_LENGTH_SHORT=0
    // Decimal point (binary point) is between bit16 and bit15

    #define FIX16_UNITY_BITPOS 16                      // # sign bit if signed int
    #define FIX16_UNITY        (1<<FIX16_UNITY_BITPOS) // 1.0000 (32768 = 0x8000) is the MOST NEGATIVE value as 1.0
                                                       //         32768 = 0x7FFF  is the MOST POSITIVE value as 1.0
                                                       // dsp_fft_forward param sine: 1 should be represented as 0x7fffffff
    #define FIX16_NEG_UNITY    (-1)                    // 11111111 all 1 as 2's complement = 0xffffffff

    typedef uint32_t ufix16_t;
    typedef int32_t  sfix16_t;

    // Based on libfixmath:
    /* Subtraction and addition no overflow detection. */

    // Not used:
    // OBSERVE that lib_dsp dsp_math.h + dsp_math.xc have all that's needed here I assume

    static inline sfix16_t fix16_add (const sfix16_t inArg0, const sfix16_t inArg1) { return (inArg0 + inArg1); }
    static inline sfix16_t fix16_sub (const sfix16_t inArg0, const sfix16_t inArg1) { return (inArg0 - inArg1); }

    // =======================================================================================
    // 16 bits = 8 bits POS and 8 bits NEG: Should make sense if MIC_ARRAY_WORD_LENGTH_SHORT=1
    // Decimal point (binary point) is between bit08 and bit07

    #define FIX08_UNITY_BITPOS 8                       // # sign bit if signed int
    #define FIX08_UNITY        (1<<FIX08_UNITY_BITPOS) // 1.0000 (256)
    #define FIX08_NEG_UNITY    (-1)                    // 11111111 all 1 as 2's complement (0xffff)

    typedef uint16_t ufix08_t;
    typedef int16_t  sfix08_t;

    // Wikipedia calls the parts "integer" and "fraction".
    // The _INT name here would cover both signed and unsigned (as said above)
    //
    #define Q_FORMAT_INT(val,num_frac_bits) (val / (1<<num_frac_bits)) // ==  (val >> num_frac_bits)

    // =======================================================================================
    // CONVERSIONS RELATED TO "dsp_math.h" in "lib_dsp". New with v0502
    // AMD Q-format: 'm' number includes sign bit, if there is one

    // All these macros work for both q8_24 (signed) and uq8_24 (unsigned)
    // Q32.0          int32_t or..
    // Q32.0         uint32_t
    //
    // Q1.31  q1_31 (  signed) = 1 bit sign,  no integer part, 31 bits fraction
    // Q1.31 uq1_31 (unsigned) =              1 bit  integer,  31 bits fraction
    //
    // Q8.24  q8_24 (  signed) = 1 bit sign,  7 bits integer,  24 bits fraction
    // Q8.24 uq8_24 (unsigned) =              8 bits integer,  24 bits fraction
    //
    //       USAGE: see q24_test_conversion_iff
    //       https://en.wikipedia.org/wiki/Q_(number_format)
    //       https://chummersone.github.io/qformat.html - Q-format Converter

                                // pow(2, 24) = 16777216
    #define UQ24_MAX 0xFFFFFFFF // 255.999999 (or 256)


    // ALL THE BELOW MAKES dsp_math_multiply_sat NEEEDED instead of dsp_math_multiply
    //
    // UNSIGNED uq8_2 OVERFLOWS SQUARED WHEN? =================== dsp_math_multiply only takes signed!
    //
    // UQ24_MAX
    // sqrt 255,999999940395355 = 15,999999998137355 times 16777216 = 268435455,.. 0x0FFFFFFF = 15,999999940395355 squared
    // 0xFFFFFFF * 0xFFFFFFF =             0xFFFFFFE0000001                        -------------==========================
    // 0xFFFFFFE0000001 >> 0x18                = 0xFFFFFFE0                                     OK HIGHEST POSSIBLE
    //                                 32bit =   0xFFFFFFE0                                     ==========================
    //                                             NNffffff  N and f 4 bits
    // 15 * 15 =
    // 0xF000000 * 0xF000000 =             0xE1000000000000                        0x10000000 = 16 squared
    // 0xE1000000000000 >> 0x18 =                0xE1000000 225 FACIT              -------------===========================
    //                                 32bit =   0xFFFFFF00                                     OK HIGHEST POSSIBLE integer
    //                                             NNffffff N and f 4 bits                      ===========================

    // SIGNED uq8_2 OVERFLOWS SQUARED WHEN? ===================== dsp_math_multiply takes signed!
    //
    // Ref MAX_Q8_24 (0x7FFFFFFF) 127.9999999..
    // 127.9999999.. * 127.9999999.. =
    // 0x7FFFFFFF * 0x7FFFFFFF =         0x3FFFFFFF00000001
    // 0x3FFFFFFF00000001 >> 0x18 =            0x3FFFFFFF00 16383,999984741210938 FACIT
    //                                 32bit =   0xFFFFFF00    255,999984741210938 OVERFLOWED, ALSO WITH SIGNED BIT FILLED!
    //                                             nNffffff n 3 bits, 1 sign bit
    //
    // sqrt 127,999999940395355.. = 11,313708496350582 times 16777216 = 189812531,.. =  0xB504F33 = 11,313708484172823 squared
    // 0xB504F33 * 0xB504F33 =             0x7FFFFFFA608429                             ------------==========================
    // 0x7FFFFFFA608429 >> 0x18 =                0x7FFFFFFA                                         OK HIGHEST POSSIBLE
    //                                             nNffffff n 3 bits, 1 sign bit                    ==========================
    //
    // 11 * 11 =                                                                                    0xB000000   11 squared
    // 0xB000000 * 0xB000000 =             0x79000000000000                                         ==========
    // 0x79000000000000 >> 0x18 =                0x79000000 121 FACIT                               OK HIGHEST POSSIBLE INTEGER
    //                                             nNffffff n 3 bits, 1 sign bit                    ===========================

    #define UQ24_128 0x80000000 // Since Q24(128) does not reach this far (is 0)
    #define UQ24_MIN 0x0        // 0.0
    //
    #define MAX_Q8_24_IF_POSITIVE_NUM_BITS 31 // Unsigned value in signed only has 31 bits
    
    #define Q24_NaN 1 // A small number both for q8_24 (signed) and uq8_24 (unsigned)
    // Used for 
    //     like log(0) or log (negative) of any log base are "Not a number"
    //     Not using: ((signed)HUGE_VAL) from "math.h" since  Error: Undefined reference to '__infinity.globound'

    // log(x) = ln(x) * log(2.71828) = ln(x) * 0.434294481903252
    // 0,434294481903252 * 2exp24 = 7286252,.. = 0x6F2DEC
    #define Q24_LOG10_OF_E_NUM 0x6F2DEC // Becomes 0.434294. See below
    //
    // Same if macros from dsp_qformat.h are used (no runtime added):
    #define Q24_LOG10_OF_E_NUM_ Q24(0.434294) // Becomes 0.434293. See above
    // Alternatively:
    //     #define BP 24   // location of the binary point
    //     Q(BP)(0.434294) // Use
    
    // From test_ln_of_1 where ln(1/(2exp24)) is found = ln(1 as uint32)
    #define Q24_LN_MIN     (-279097920) //             -279097920 printed out
    #define Q24_LN_MIN_ Q24(-16.635532) // resolves to -279097914 Q24 macro calculated

    #define Q_FORMAT_STR_LEN 12 // "-128.123456" space for \0 needed
                                //  123456789te
 
    signed q24_signed_to_str (           // returns num_chars
            char          chars [],      // Includes \0 character
            const int32_t value_signed); // always signed
    
    signed uq24_unsigned_to_str (           // returns num_chars
            char           chars [],        // Includes \0 character
            const uint32_t value_unsigned); // always unsigned

    typedef struct {
        unsigned err_cnt;
        unsigned debug_tag;
    } error_info_t;

    typedef struct {
        uint32_t num; // numerator
        uint32_t den; // denominator
    } uint32_ratio_t;

    typedef struct {
        int32_t num; // numerator
        int32_t den; // denominator
    } int32_ratio_t;

    // Called if DO_Q24_SET_DEBUG_TAG==1
    //
    bool set_debug_tag ( // returns true if triggered condition inside
              error_info_t       &error, // io
              const error_info_t error_prev,
              const unsigned     tag_bitpos, // 0 to avoid DO_Q24_DEBUG_PRINT_IF_BITPOS
              const signed       value);

    // If we need to know about when dsp_math_multiply_sat saturated
    //
    int32_t my_dsp_math_multiply_sat ( // returns product_32_sat
            int32_t       input1_value,
            int32_t       input2_value,
            const int32_t q_format,
            error_info_t  &error);

    // Multiply not allowed to overflow/underflow into highest 32 bits word
    //
    int32_t my_math_multiply_32bit_sat ( // returns product_32_sat
            const int32_t value, // Basically Q1.31
            const int32_t mul_N, // --"--
            error_info_t  &error);

    // Calculate dB (decibel) of unsigned uq8_24
    //
    q8_24 my_dB (
            const uq8_24 operand,
            error_info_t &error);

    unsigned find_downscale_by_bits_param (
        const unsigned max_range_num_bits,
        const unsigned scaled_to_num_bits,
        unsigned       val_positive); // positive but packed in unsigned (for "math_dsp reasons")

    uint32_t downscale_by_uint32_ratio (
            const uint32_t       value,
            const uint32_ratio_t ratio);

    // =======================================
    // Understanding som dsp_math.h functions
    // =======================================
    //
    //     q8_24 dB = 0;
    //     THIS:
    //         dB = dsp_math_multiply (dB, Q24_LOG10_OF_E_NUM, 24);                    // 1.53604
    //     OR THIS cast to 64 bit MUST be done on EACH of the values:
    //         dB = (q8_24) ((((uint64_t) dB * (uint64_t) Q24_LOG10_OF_E_NUM)) >> 24); // 1.53604
    
    //      in math_fix.xc
    
    #ifndef DO_Q24_DEBUG_PRINT_IF_MULT_SATURATED
        #error include of _globals.h missing
    #endif

    #ifndef DO_Q24_SET_DEBUG_TAG
        #error include of _globals.h missing
    #endif
    
    #ifndef DO_Q24_TEST_MATH
        #error include of _globals.h missing
    #elif (DO_Q24_TEST_MATH==1)
        void q24_test_math_iff (void);
    #else
        #define q24_test_math_iff()
    #endif
    
    #ifndef DO_Q24_TEST_CONVERSION
        #error include of _globals.h missing
    #elif (DO_Q24_TEST_CONVERSION==1)
        void q24_test_conversion_iff (void);
    #else
        #define q24_test_conversion_iff()
    #endif

#else
    // Requires
    // #include <dsp_math.h>
    // #include <dsp_qformat.h>
    #error NESTED .h FILE
#endif
